use std::{
	env::current_dir,
	io::{ stdout, IsTerminal },
	path::Path,
	process::{ exit, Command, Stdio }
};

use pico_args::Arguments;
use toml::value::Value;

mod config;
mod error;
mod util;

use config::Config;

fn main() {
	let mut args = Arguments::from_env();

	//	help flag		(-h / --help)
	if args.contains(["-h", "--help"]) {
		help_text();
		return;
	}
	//	version flag	(-v / --version)
	if args.contains(["-v", "--version"]) {
		println!("{}", env!("CARGO_PKG_VERSION"));
		return;
	}

	//	prepare configs
	let i_dir = current_dir().unwrap();
	let dir = i_dir.as_path();
	let config = Config::new();

	//	path flag		(-p / --path)
	if args.contains(["-p", "--path"]) {
		if let Some(local) = config.local_path { println!("{local}"); }
		else if let Some(global) = config.global_path { println!("{global}"); }
		return;
	}

	//	get target
	let arg_target = args.subcommand().unwrap();
	let i_target = arg_target.unwrap_or(String::from("."));
	let target = Path::new(&i_target);
	if !target.exists() { error::not_found(&target); }

	//	get section
	//	ordering: filename -> type (ext/dir) -> default
	let mut section = None;

	//	by exact filename
	if let Some(filename) = target.file_name() {
		let array = config.get("filename");
		let matches: Vec<Value>;
		if let Some(Value::Array(array)) = array { matches = util::matches(array, filename.to_string_lossy().into()); }
		else { matches = Vec::new(); }

		section = if matches.len() > 0 { matches.get(0).cloned() } else { None };
	}

	//	handle types; dir first
	if section.is_none() && target.is_dir() {
		let dir_section = config.get("dir");
		if dir_section.is_some() {
			section = dir_section;
		}
	}

	//	handle types; extensions second
	if section.is_none() {
		if let Some(extension) = target.extension() {
			let array: Option<Value> = config.get("extension");
			let matches: Vec<Value>;
			if let Some(Value::Array(array)) = array { matches = util::matches(array, extension.to_string_lossy().into()); }
			else { matches = Vec::new(); }

			section = if matches.len() > 0 { matches.get(0).cloned() } else { None };
		}
	}

	//	default or fail on missing session
	if section.is_none() {
		let default = config.get("default");
		if default.is_some() { section = default; }
		else { error::no_section(target); }
	}

	//	unwrap our section
	let properties = section.unwrap();

	//	collect properties
	let command = properties.get("command").unwrap().as_str().unwrap().to_string();
	let shell = properties.get("shell").unwrap_or(&Value::Boolean(false)).as_bool().unwrap();

	if !stdout().is_terminal() {
		println!("{command} {i_target}");
		exit(0);
	}

	//	build child
	let mut parts = command.split(" ");
	let mut child = Command::new(parts.next().unwrap());
	let mut param: Vec<&str> = parts.collect();
	param.push(&i_target);
	child
		.args(param)
		.current_dir(dir);

	//	shell processes inherit stdio (and use `output()`)
	if shell {
		child
			.stdin(Stdio::inherit())
			.stdout(Stdio::inherit())
			.stderr(Stdio::inherit());
		child.output().ok();
	}
	//	non-shell processes should redirect to dev/null (and use `spawn()`)
	else {
		child
			.stdin(Stdio::null())
			.stdout(Stdio::null())
			.stderr(Stdio::null());
		child.spawn().ok();
	}
}

pub fn help_text() {
	println!("open v{}
Valerie Wolfe <sleeplessval@gmail.com>
A Rust reimagining of \"xdg-open\" configurable with an toml file.

usage: open [flags] <target>

flags:
   -h, --help      Prints this help text
   -p, --path      Prints the path to the config file
   -v, --version   Prints the version number
",
	env!("CARGO_PKG_VERSION")
	);
}

